/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 *   Jan Varga <varga@utcru.sk> (Original Author)
 */

#include "prprf.h"

#include "nsString.h"
#include "nsIServiceManager.h"
#include "nsXPIDLString.h"
#include "nsRDFCID.h"
#include "rdf.h"

#include "nsDatabaseDataSource.h"
#include "nsIDatabaseResource.h"
#include "nsIDatabaseRow.h"


#define DATABASE_URI "database://"


static nsIRDFService* gRDFService = nsnull;


NS_IMPL_ISUPPORTS1(nsResultEnumerator, nsISimpleEnumerator);

nsResultEnumerator::nsResultEnumerator(PRInt32 aRowCount)
{
  NS_INIT_ISUPPORTS();
  mRowIndex = 0;
  mRowCount = aRowCount;
}

nsResultEnumerator::~nsResultEnumerator()
{
}

NS_IMETHODIMP nsResultEnumerator::HasMoreElements(PRBool* _retval)
{                                                                                               
  *_retval = (mRowIndex < mRowCount);
  
  return NS_OK;
}

NS_IMETHODIMP nsResultEnumerator::GetNext(nsISupports** _retval)
{
  char uri[32];
  PR_snprintf(uri, 32, DATABASE_URI "%d", mRowIndex);
  nsCOMPtr<nsIRDFResource> resource;
  gRDFService->GetResource(uri, getter_AddRefs(resource));
  nsCOMPtr<nsIDatabaseResource> databaseResource = do_QueryInterface(resource);
  databaseResource->SetRowIndex(mRowIndex);
  mRowIndex++;

  *_retval = resource;
  NS_ADDREF(*_retval);
  
  return NS_OK;
}


static NS_DEFINE_CID(kRDFServiceCID, NS_RDFSERVICE_CID);

PRInt32             nsDatabaseDataSource::gRefCnt = 0;
nsIRDFResource*     nsDatabaseDataSource::kNC_Child;
nsIRDFResource*     nsDatabaseDataSource::kDatabaseRoot;


nsDatabaseDataSource::nsDatabaseDataSource(void)
{
#ifdef DEBUG
  printf("nsDatabaseDataSource::nsDatabaseDataSource\n");
#endif
  NS_INIT_ISUPPORTS();

  if (gRefCnt++ == 0) {
    nsresult rv;
    rv = nsServiceManager::GetService(kRDFServiceCID,
                                      NS_GET_IID(nsIRDFService),
                                      NS_REINTERPRET_CAST(nsISupports**, &gRDFService));

    gRDFService->GetResource(NC_NAMESPACE_URI "child", &kNC_Child);
    gRDFService->GetResource(DATABASE_URI, &kDatabaseRoot);
  }
}

nsDatabaseDataSource::~nsDatabaseDataSource (void)
{
#ifdef DEBUG
  printf("nsDatabaseDataSource::~nsDatabaseDataSource\n");
#endif
    
  if (--gRefCnt == 0) {
    NS_RELEASE(kDatabaseRoot);
    NS_RELEASE(kNC_Child);
    nsServiceManager::ReleaseService(kRDFServiceCID, gRDFService);
    gRDFService = nsnull;
  }
}


NS_IMPL_ISUPPORTS2(nsDatabaseDataSource,
                   nsIRDFDataSource,
                   nsIDatabaseDataSource);


NS_IMETHODIMP
nsDatabaseDataSource::GetURI(char **aURI)
{
  if (!aURI)
    return NS_ERROR_NULL_POINTER;

  *aURI = nsXPIDLCString::Copy("rdf:database");
  if (! *aURI )
    return NS_ERROR_OUT_OF_MEMORY;
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetSource(nsIRDFResource* aPoperty,
                                nsIRDFNode* aTarget,
                                PRBool aTruthValue,
                                nsIRDFResource** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetSources(nsIRDFResource* aProperty,
                                 nsIRDFNode* aTarget,
                                 PRBool aTruthValue,
                                 nsISimpleEnumerator** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetTarget(nsIRDFResource* aSource,
                                nsIRDFResource* aProperty,
                                PRBool aTruthValue,
                                nsIRDFNode** _retval)
{
  PRInt32 rowIndex;
  nsCOMPtr<nsIDatabaseResource> databaseResource = do_QueryInterface(aSource);
  if (! databaseResource)
    return NS_RDF_NO_VALUE;
  databaseResource->GetRowIndex(&rowIndex);

  PRInt16 columnIndex;
  const char* valueConst;
  aProperty->GetValueConst(&valueConst);
  mRowList->GetColumnIndex(NS_ConvertASCIItoUCS2(valueConst).get(), &columnIndex);
  if (columnIndex < 0)
    return NS_RDF_NO_VALUE;

  nsCOMPtr<nsIDatabaseRow> row;
  mRowList->GetRow(rowIndex, getter_AddRefs(row));
  const PRUnichar* value;
  row->GetValueConst(columnIndex, &value);
  if (! value)
    return NS_RDF_NO_VALUE;
  nsCOMPtr<nsIRDFLiteral> literal;
  gRDFService->GetLiteral(value, getter_AddRefs(literal));
  if (! literal)
    return NS_RDF_NO_VALUE;
  
  *_retval = literal;
  NS_ADDREF(*_retval);
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetTargets(nsIRDFResource* aSource,
                                 nsIRDFResource* aProperty,
                                 PRBool aTruthValue,
                                 nsISimpleEnumerator** _retval)
{
  if (aSource == kDatabaseRoot &&
      aProperty == kNC_Child &&
      aTruthValue &&
      mRowList) {
    PRInt32 rowCount;
    mRowList->GetRowCount(&rowCount);
    nsISimpleEnumerator* enumerator = new nsResultEnumerator(rowCount);
    if (! enumerator)
      return NS_ERROR_OUT_OF_MEMORY;
    NS_ADDREF(enumerator);
    *_retval = enumerator;    
    
    return NS_OK;
  }    
  else
    return NS_RDF_NO_VALUE;
}

NS_IMETHODIMP
nsDatabaseDataSource::Assert(nsIRDFResource* aSource,
                             nsIRDFResource* aProperty,
                             nsIRDFNode* aTarget,
                             PRBool aTruthValue)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}


NS_IMETHODIMP
nsDatabaseDataSource::Unassert(nsIRDFResource* aSource,
                               nsIRDFResource* aProperty,
                               nsIRDFNode* aTarget)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::Change(nsIRDFResource* aSource,
                             nsIRDFResource* aProperty,
                             nsIRDFNode* aOldTarget,
                             nsIRDFNode* aNewTarget)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::Move(nsIRDFResource* aOldSource,
                           nsIRDFResource* aNewSource,
                           nsIRDFResource* aProperty,
                           nsIRDFNode* aTarget)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::HasAssertion(nsIRDFResource* aSource,
                                   nsIRDFResource* aProperty,
                                   nsIRDFNode *aTarget,
                                   PRBool aTruthValue,
                                   PRBool* _retval)
{
  *_retval = PR_FALSE;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::AddObserver(nsIRDFObserver *aObserver)
{
  if (! aObserver)
    return NS_ERROR_NULL_POINTER;

  if (! mObservers) {
    nsresult rv;
    rv = NS_NewISupportsArray(getter_AddRefs(mObservers));
    if (NS_FAILED(rv)) return rv;
  }
  
  mObservers->AppendElement(aObserver);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::RemoveObserver(nsIRDFObserver *aObserver)
{
  if (! aObserver)
    return NS_ERROR_NULL_POINTER;

  if (! mObservers)
    return NS_OK;

  mObservers->RemoveElement(aObserver);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::HasArcIn(nsIRDFNode* aNode,
                               nsIRDFResource* aArc,
                               PRBool* _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}                                                                                       

NS_IMETHODIMP
nsDatabaseDataSource::HasArcOut(nsIRDFResource* aSource,
                                nsIRDFResource*aArc,
                                PRBool* _retval)
{                                                                                               
  *_retval = PR_FALSE;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::ArcLabelsIn(nsIRDFNode* aTarget,
                                  nsISimpleEnumerator** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::ArcLabelsOut(nsIRDFResource* aSource,
                                   nsISimpleEnumerator** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetAllResources(nsISimpleEnumerator** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}


NS_IMETHODIMP
nsDatabaseDataSource::GetAllCommands(nsIRDFResource* aSource,
                                     nsIEnumerator** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetAllCmds(nsIRDFResource* aSource,
                                 nsISimpleEnumerator** _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::IsCommandEnabled(nsISupportsArray* aSources,
                                       nsIRDFResource* aCommand,
                                       nsISupportsArray* aArguments,
                                       PRBool* _retval)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::DoCommand(nsISupportsArray* aSources,
                                nsIRDFResource* aCommand,
                                nsISupportsArray* aArguments)
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseDataSource::GetRowList(nsIDatabaseRowList** aRowList)
{
  *aRowList = mRowList;
  NS_IF_ADDREF(*aRowList);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseDataSource::SetRowList(nsIDatabaseRowList* aRowList)
{
  mRowList = aRowList;

  return NS_OK;
}
