/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 *   Jan Varga <varga@utcru.sk> (Original Author)
 */

#include "nsMemory.h"
#include "nsCRT.h"
#include "nsXPIDLString.h"

#include "nsDatabaseRowList.h"
#include "nsDatabaseRow.h"


//#define DEBUG 1
#define DEBUG_REFCOUNT 1


#ifdef DEBUG_REFCOUNT
static PRInt32 gCount = 0;
#endif


nsDatabaseRowList::nsDatabaseRowList()
{
#ifdef DEBUG
  printf("nsDatabaseRowList::nsDatabaseRowList\n");
#endif
#ifdef DEBUG_REFCOUNT
  gCount++;
#endif
  NS_INIT_ISUPPORTS();
  mColumnCount = 0;
  mColumnNames = nsnull;
}

nsDatabaseRowList::~nsDatabaseRowList()
{
#ifdef DEBUG
  printf("nsDatabaseRowList::~nsDatabaseRowList\n");
#endif
#ifdef DEBUG_REFCOUNT
  if (--gCount == 0)
    printf("Last row list\n");
#endif
  for (PRInt32 i = mRows.Count() - 1; i >= 0; i--) {
    nsIDatabaseRow* row = NS_STATIC_CAST(nsIDatabaseRow*, mRows.ElementAt(i));
    NS_RELEASE(row);
  }

  if (mColumnNames) {
    for (PRInt16 i = 0; i < mColumnCount; i++)
      if (mColumnNames[i])
        nsMemory::Free(mColumnNames[i]);
    nsMemory::Free(mColumnNames);
  }
}


NS_IMPL_ISUPPORTS1(nsDatabaseRowList,
                   nsIDatabaseRowList);


NS_IMETHODIMP
nsDatabaseRowList::Init(PRInt16 aColumnCount)
{
  mColumnCount = aColumnCount;
  PRInt32 bytes = mColumnCount * sizeof(PRUnichar*);
  mColumnNames = (PRUnichar**)nsMemory::Alloc(bytes);
  nsCRT::memset(mColumnNames, 0, bytes);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::GetRowCount(PRInt32* aRowCount)
{
  *aRowCount = mRows.Count();

  return NS_OK;
}
    
NS_IMETHODIMP
nsDatabaseRowList::GetColumnCount(PRInt16* aColumnCount)
{
  *aColumnCount = mColumnCount;
  
  return NS_OK;
}
         
NS_IMETHODIMP
nsDatabaseRowList::GetColumnName(PRInt16 aColumnIndex, PRUnichar** _retval)
{
  if (! mColumnNames)
    return NS_ERROR_NOT_INITIALIZED;
    
  if (aColumnIndex >= 0 && aColumnIndex < mColumnCount && mColumnNames[aColumnIndex])
    *_retval = nsXPIDLString::Copy(mColumnNames[aColumnIndex]);
  else
    *_retval = nsnull;
  
  return NS_OK;
}
    
NS_IMETHODIMP
nsDatabaseRowList::SetColumnName(PRInt16 aColumnIndex, const PRUnichar* aColumnName)
{
  if (! mColumnNames)
    return NS_ERROR_NOT_INITIALIZED;

  if (aColumnIndex >= 0 && aColumnIndex < mColumnCount) {
    if (mColumnNames[aColumnIndex])
      nsMemory::Free(mColumnNames[aColumnIndex]);
    mColumnNames[aColumnIndex] = nsXPIDLString::Copy(aColumnName);
  }
  
  return NS_OK;
}
    
NS_IMETHODIMP
nsDatabaseRowList::GetColumnIndex(const PRUnichar* aColumnName, PRInt16* _retval)
{
  if (! mColumnNames)
    return NS_ERROR_NOT_INITIALIZED;

  *_retval = -1;

  for (PRInt16 i = 0; i < mColumnCount; i++)
    if (mColumnNames[i])
      if (!nsCRT::strcmp(aColumnName, mColumnNames[i])) {
        *_retval = i;
        break;
      }
  
  return NS_OK;
}
        
NS_IMETHODIMP
nsDatabaseRowList::CreateRow(nsIDatabaseRow** _retval)
{
  nsIDatabaseRow* row;
  row = new nsDatabaseRow();
  row->Init(mColumnCount);
  *_retval = row;
  NS_ADDREF(*_retval);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::GetRow(PRInt32 aRowIndex, nsIDatabaseRow** _retval)
{
  if (aRowIndex >= 0 && aRowIndex < mRows.Count()) {
    *_retval = NS_STATIC_CAST(nsIDatabaseRow*, mRows.ElementAt(aRowIndex));
    NS_ADDREF(*_retval);
  }
  else
    *_retval = nsnull;
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::AppendRow(nsIDatabaseRow *aRow)
{
  if (mObservers) {
     PRUint32 count;
     mObservers->Count(&count);
     for (PRUint32 i = 0; i < count; i++) {
       nsCOMPtr<nsIDatabaseRowListObserver> observer;
       mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseRowListObserver), getter_AddRefs(observer));
       if (observer)
         observer->OnAppendRow(aRow);
     }
  }

  mRows.AppendElement(aRow);
  NS_ADDREF(aRow);

  return NS_OK;
}
            
NS_IMETHODIMP
nsDatabaseRowList::InsertRow(PRInt32 aRowIndex, nsIDatabaseRow *aRow)
{
  if (aRowIndex >= 0 && aRowIndex < mRows.Count()) {
    if (mObservers) {
       PRUint32 count;
       mObservers->Count(&count);
       for (PRUint32 i = 0; i < count; i++) {
         nsCOMPtr<nsIDatabaseRowListObserver> observer;
         mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseRowListObserver), getter_AddRefs(observer));
         if (observer)
           observer->OnInsertRow(aRowIndex, aRow);
       }
    }
 
    mRows.InsertElementAt(aRow, aRowIndex);
    NS_ADDREF(aRow);
  }
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::ReplaceRow(PRInt32 aRowIndex, nsIDatabaseRow *aRow)
{
  if (aRowIndex >= 0 && aRowIndex < mRows.Count()) {
    if (mObservers) {
       PRUint32 count;
       mObservers->Count(&count);
       for (PRUint32 i = 0; i < count; i++) {
         nsCOMPtr<nsIDatabaseRowListObserver> observer;
         mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseRowListObserver), getter_AddRefs(observer));
         if (observer)
           observer->OnReplaceRow(aRowIndex, aRow);
       }
    }

    nsIDatabaseRow* oldRow = NS_STATIC_CAST(nsIDatabaseRow*, mRows.ElementAt(aRowIndex));
    NS_RELEASE(oldRow);

    mRows.ReplaceElementAt(aRow, aRowIndex);
    NS_ADDREF(aRow);
  }
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::RemoveRow(PRInt32 aRowIndex)
{
  if (aRowIndex >= 0 && aRowIndex < mRows.Count()) {
    if (mObservers) {
       PRUint32 count;
       mObservers->Count(&count);
       for (PRUint32 i = 0; i < count; i++) {
         nsCOMPtr<nsIDatabaseRowListObserver> observer;
         mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseRowListObserver), getter_AddRefs(observer));
         if (observer)
           observer->OnRemoveRow(aRowIndex);
       }
    }

    nsIDatabaseRow* oldRow = NS_STATIC_CAST(nsIDatabaseRow*, mRows.ElementAt(aRowIndex));
    NS_RELEASE(oldRow);
  
    mRows.RemoveElementAt(aRowIndex);
  }
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::ClearRows()
{
  if (mObservers) {
     PRUint32 count;
     mObservers->Count(&count);
     for (PRUint32 i = 0; i < count; i++) {
       nsCOMPtr<nsIDatabaseRowListObserver> observer;
       mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseRowListObserver), getter_AddRefs(observer));
       if (observer)
         observer->OnClearRows();
     }
  }

  for (PRInt32 i = mRows.Count() - 1; i >= 0; i--) {
    nsIDatabaseRow* row = NS_STATIC_CAST(nsIDatabaseRow*, mRows.ElementAt(i));
    NS_RELEASE(row);
  }

  mRows.Clear();
  mRows.Compact();

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::AddObserver(nsIDatabaseRowListObserver* aObserver)
{
  if (! mObservers) {
    nsresult rv;
    rv = NS_NewISupportsArray(getter_AddRefs(mObservers));
    if (NS_FAILED(rv))
      return rv;
  }
                
  mObservers->AppendElement(aObserver);
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseRowList::RemoveObserver(nsIDatabaseRowListObserver* aObserver)
{
  if (! mObservers)
    return NS_OK;
        
  mObservers->RemoveElement(aObserver);
  return NS_OK;
}
