/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 *   Jan Varga <varga@utcru.sk> (Original Author)
 */

#include "prprf.h"

#include "nsString.h"
#include "nsReadableUtils.h"

#include "nsDatabaseConnectionMysql.h"
#include "nsDatabaseCID.h"


#define DEBUG 1


static NS_DEFINE_CID(kProxyObjectManagerCID, NS_PROXYEVENT_MANAGER_CID);
static NS_DEFINE_CID(kDatabaseRowListCID, NS_DATABASE_ROW_LIST_CID);


nsDatabaseConnectionMysql::nsDatabaseConnectionMysql()
{
#ifdef DEBUG
  printf("nsDatabaseConnectionMysql::nsDatabaseConnectionMysql\n");
#endif
  NS_INIT_ISUPPORTS();
  mStatusCode = ERR_OK;
  mStatusMessage.Assign("");
  mWindow = nsnull;
  mConnection = 0;
  mConnected = PR_FALSE;
  mLock = PR_NewLock();
  mCondVar = PR_NewCondVar(mLock);
  mRunning = PR_FALSE;
  mCancelled = PR_FALSE;

  mProxyObjectManager = do_GetService(kProxyObjectManagerCID);

  NS_NewThread(getter_AddRefs(mThread), this, 0, PR_UNJOINABLE_THREAD);
}

nsDatabaseConnectionMysql::~nsDatabaseConnectionMysql()
{
#ifdef DEBUG
  printf("nsDatabaseConnectionMysql::~nsDatabaseConnectionMysql\n");
#endif
  if (mConnection)
    mysql_close(mConnection);
  PR_DestroyCondVar(mCondVar);
  PR_DestroyLock(mLock);
}


NS_IMPL_ISUPPORTS3(nsDatabaseConnectionMysql,
                   nsIRunnable,
                   nsIDatabaseConnection,
                   nsIDatabaseConnectionMysql)


NS_IMETHODIMP
nsDatabaseConnectionMysql::Run()
{
  while (1) {
    PR_Lock(mLock);
    PR_WaitCondVar(mCondVar, PR_INTERVAL_NO_TIMEOUT);

    mRunning = PR_TRUE;

    mCancelled = PR_FALSE;

    mWindow->SetCursor(NS_LITERAL_STRING("wait"));

    NotifyObservers(MSG_BEGIN_EXECUTION, PR_TRUE);

    mysql_query(mConnection, NS_ConvertUCS2toUTF8(mQuery).get());
    if (mysql_errno(mConnection)) {
      mListener->OnError(this, mContext, ERR_QUERY_FAILED, mysql_error(mConnection));
    }
    else {
      MYSQL_RES* result = mysql_store_result(mConnection);
      nsCOMPtr<nsIDatabaseRowList> rowList;
      BuildRows(result, getter_AddRefs(rowList));
      mysql_free_result(result);
      mListener->OnResult(this, mContext, rowList);
    }

    mListener = nsnull;

    mContext = nsnull;

    NotifyObservers(MSG_END_EXECUTION, PR_TRUE);

    mWindow->SetCursor(NS_LITERAL_STRING("default"));

    mRunning = PR_FALSE;

    PR_Unlock(mLock);
  }

  return NS_OK;
}


NS_IMETHODIMP
nsDatabaseConnectionMysql::GetStatusCode(PRInt32* aStatusCode)
{
  *aStatusCode = mStatusCode;
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::GetStatusMessage(char** aStatusMessage)
{
  *aStatusMessage = ToNewCString(mStatusMessage);
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::GetWindow(nsIDOMWindowInternal** aWindow)
{
  *aWindow = mWindow;
  NS_IF_ADDREF(*aWindow);

  return NS_OK;
}

NS_IMETHODIMP nsDatabaseConnectionMysql::Init(nsIDOMWindowInternal* aWindow)
{
  mWindow = aWindow;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::Connect(const char* aHost, PRInt32 aPort, const char* aDatabase, const char* aUsername, const char* aPassword)
{
  if (mConnection) {
    mStatusCode = ERR_CONNECTED;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  mConnection = mysql_init(0);

  mysql_real_connect(mConnection, aHost, aUsername, aPassword, aDatabase, aPort, 0, 0);
  if (mysql_errno(mConnection)) {
    mStatusCode = ERR_CONNECT_FAILED;
    mStatusMessage.Assign(mysql_error(mConnection));
    mConnection = nsnull;
  }
  else {
    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
    mConnected = PR_TRUE;
  }

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::GetConnected(PRBool *aConnected)
{
  *aConnected = mConnected;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::Disconnect()
{
  if (! mConnection) {
    mStatusCode = ERR_NOT_CONNECTED;
    mStatusMessage.Assign("");
  }
  else {
    mysql_close(mConnection);

    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
    mConnected = PR_FALSE;
    mConnection = nsnull;
  }

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::GetRunning(PRBool *aRunning)
{
  *aRunning = mRunning;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::Exec(const PRUnichar* aQuery, nsIDatabaseRowList** _retval)
{
  *_retval = nsnull;

  if (! mWindow)
    return NS_ERROR_NOT_INITIALIZED;

  if (! mConnection) {
    mStatusCode = ERR_NOT_CONNECTED;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  if (mRunning) {
    mStatusCode = ERR_RUNNING;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  PR_Lock(mLock);

  mRunning = PR_TRUE;

  mCancelled = PR_FALSE;

  mWindow->SetCursor(NS_LITERAL_STRING("wait"));

  NotifyObservers(MSG_BEGIN_EXECUTION, PR_FALSE);

  mysql_query(mConnection, NS_ConvertUCS2toUTF8(aQuery).get());
  if (mysql_errno(mConnection)) {
    mStatusCode = ERR_QUERY_FAILED;
    mStatusMessage.Assign(mysql_error(mConnection));
  }
  else {
    MYSQL_RES* result = mysql_store_result(mConnection);
    BuildRows(result, _retval);
    mysql_free_result(result);
    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
  }

  NotifyObservers(MSG_END_EXECUTION, PR_FALSE);

  mWindow->SetCursor(NS_LITERAL_STRING("default"));

  mRunning = PR_FALSE;

  PR_Unlock(mLock);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::AsyncExec(const PRUnichar *aQuery, nsIDatabaseConnectionListener* aListener, nsISupports* aContext)
{
  if (! aListener)
    return NS_ERROR_NULL_POINTER;

  if (! mWindow)
    return NS_ERROR_NOT_INITIALIZED;

  if (! mConnection) {
    aListener->OnError(this, aContext, ERR_NOT_CONNECTED, "");
    return NS_OK;
  }

  if (mRunning) {
    aListener->OnError(this, aContext, ERR_RUNNING, "");
    return NS_OK;
  }

  mQuery.Assign(aQuery);

  mProxyObjectManager->GetProxyForObject(NS_UI_THREAD_EVENTQ,
                                         NS_GET_IID(nsIDatabaseConnectionListener),
                                         aListener,
                                         PROXY_SYNC | PROXY_ALWAYS,
                                         getter_AddRefs(mListener));

  mContext = aContext;

  PR_Lock(mLock);
  PR_NotifyCondVar(mCondVar);
  PR_Unlock(mLock);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::GetCancelled(PRBool *aCancelled)
{
  *aCancelled = mCancelled;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::Stop()
{
  return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
nsDatabaseConnectionMysql::AddObserver(nsIDatabaseConnectionObserver* aObserver)
{
  if (! mObservers) {
    nsresult rv;
    rv = NS_NewISupportsArray(getter_AddRefs(mObservers));
    if (NS_FAILED(rv))
      return rv;
  }
              
  mObservers->AppendElement(aObserver);
            
  return NS_OK;
}
               
NS_IMETHODIMP
nsDatabaseConnectionMysql::RemoveObserver(nsIDatabaseConnectionObserver* aObserver)
{
  if (! mObservers)
    return NS_OK;
      
  mObservers->RemoveElement(aObserver);
    
  return NS_OK;
}

      
nsresult
nsDatabaseConnectionMysql::BuildRows(MYSQL_RES* aResult, nsIDatabaseRowList** _retval)
{
#ifdef DEBUG
  PRIntervalTime startTime =  PR_IntervalNow();
#endif

  PRInt32 columnCount = mysql_num_fields(aResult);

  nsCOMPtr<nsIDatabaseRowList> rowList;
  nsComponentManager::CreateInstance(kDatabaseRowListCID,
                                     nsnull,
                                     NS_GET_IID(nsIDatabaseRowList),
                                     getter_AddRefs(rowList));

  rowList->Init(columnCount);
  MYSQL_FIELD* fields = mysql_fetch_fields(aResult);
  PRInt32 i;
  for (i = 0; i < columnCount; i++)
    rowList->SetColumnName(i, NS_ConvertUTF8toUCS2(fields[i].name).get());
      
  for (i = 0; i < mysql_num_rows(aResult); i++) {
    nsCOMPtr<nsIDatabaseRow> databaseRow;
    rowList->CreateRow(getter_AddRefs(databaseRow));
    MYSQL_ROW row = mysql_fetch_row(aResult);
    PRInt32 j;
    for (j = 0; j < columnCount; j++)
      databaseRow->SetValue(j, NS_ConvertUTF8toUCS2(row[j]).get());
    rowList->AppendRow(databaseRow);
  }

#ifdef DEBUG
  PRIntervalTime endTime =  PR_IntervalNow();
  printf("Time spent with building rows: %d msec\n", PR_IntervalToMilliseconds(endTime - startTime));
#endif

  *_retval = rowList;
  NS_ADDREF(*_retval);
  
  return NS_OK;
}

nsresult
nsDatabaseConnectionMysql::NotifyObservers(char aMsgType, PRBool aUseProxy)
{
  if (mObservers) {
     PRUint32 count;
     mObservers->Count(&count);
     for (PRUint32 i = 0; i < count; i++) {
       nsCOMPtr<nsIDatabaseConnectionObserver> observer;
       mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseConnectionObserver), getter_AddRefs(observer));
       if (observer) {
         nsCOMPtr<nsIDatabaseConnectionObserver> realObserver;
         if (aUseProxy)
           mProxyObjectManager->GetProxyForObject(NS_UI_THREAD_EVENTQ,
                                                  NS_GET_IID(nsIDatabaseConnectionObserver),
                                                  observer,
                                                  PROXY_SYNC | PROXY_ALWAYS,
                                                  getter_AddRefs(realObserver));
         else
           realObserver = observer;
       
         if (aMsgType == MSG_BEGIN_EXECUTION)
           realObserver->OnBeginExecution();
         else if (aMsgType == MSG_END_EXECUTION)
           realObserver->OnEndExecution();
       }
     }
  }

  return NS_OK;
}
