/*
 * The contents of this file are subject to the Mozilla Public
 * License Version 1.1 (the "License"); you may not use this file
 * except in compliance with the License. You may obtain a copy of
 * the License at http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS
 * IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
 * implied. See the License for the specific language governing
 * rights and limitations under the License.
 * 
 * The Original Code is mozilla.org code.
 *
 * The Initial Developer of the Original Code is Netscape
 * Communications Corporation.  Portions created by Netscape are
 * Copyright (C) 1998 Netscape Communications Corporation. All
 * Rights Reserved.
 *
 * Contributor(s): 
 *   Jan Varga <varga@utcru.sk> (Original Author)
 */

#include "prprf.h"

#include "nsReadableUtils.h"

#include "nsDatabaseConnectionPgsql.h"
#include "nsDatabaseCID.h"


//#define DEBUG 1


static NS_DEFINE_CID(kProxyObjectManagerCID, NS_PROXYEVENT_MANAGER_CID);
static NS_DEFINE_CID(kDatabaseRowListCID, NS_DATABASE_ROW_LIST_CID);


nsDatabaseConnectionPgsql::nsDatabaseConnectionPgsql()
{
#ifdef DEBUG
  printf("nsDatabaseConnectionPgsql::nsDatabaseConnectionPgsql\n");
#endif
  NS_INIT_ISUPPORTS();
  mStatusCode = ERR_OK;
  mStatusMessage.Assign("");
  mWindow = nsnull;
  mConnection = 0;
  mConnected = PR_FALSE;
  mLock = PR_NewLock();
  mCondVar = PR_NewCondVar(mLock);
  mRunning = PR_FALSE;
  mCancelled = PR_FALSE;

  mProxyObjectManager = do_GetService(kProxyObjectManagerCID);

  NS_NewThread(getter_AddRefs(mThread), this, 0, PR_UNJOINABLE_THREAD);
}

nsDatabaseConnectionPgsql::~nsDatabaseConnectionPgsql()
{
#ifdef DEBUG
  printf("nsDatabaseConnectionPgsql::~nsDatabaseConnectionPgsql\n");
#endif
  if ( mConnection )
    PQfinish(mConnection);
  PR_DestroyCondVar(mCondVar);
  PR_DestroyLock(mLock);
}


NS_IMPL_ISUPPORTS3(nsDatabaseConnectionPgsql,
                   nsIRunnable,
                   nsIDatabaseConnection,
                   nsIDatabaseConnectionPgsql)


NS_IMETHODIMP
nsDatabaseConnectionPgsql::Run()
{
  while (1) {
    PR_Lock(mLock);
    PR_WaitCondVar(mCondVar, PR_INTERVAL_NO_TIMEOUT);

    mRunning = PR_TRUE;

    mCancelled = PR_FALSE;

    mWindow->SetCursor(NS_LITERAL_STRING("wait"));

    NotifyObservers(MSG_BEGIN_EXECUTION, PR_TRUE);

    PGresult* result;
    result = PQexec(mConnection, NS_ConvertUCS2toUTF8(mQuery).get());

    PRInt32 stat;
    stat = PQresultStatus(result);
    if ((stat != PGRES_COMMAND_OK) && (stat != PGRES_TUPLES_OK))
      mListener->OnError(this, mContext, ERR_QUERY_FAILED, PQresultErrorMessage(result));
    else {
      nsCOMPtr<nsIDatabaseRowList> rowList;
      BuildRows(result, getter_AddRefs(rowList));
      mListener->OnResult(this, mContext, rowList);
    }

    PQclear(result);

    mListener = nsnull;

    mContext = nsnull;

    NotifyObservers(MSG_END_EXECUTION, PR_TRUE);

    mWindow->SetCursor(NS_LITERAL_STRING("default"));

    mRunning = PR_FALSE;

    PR_Unlock(mLock);
  }

  return NS_OK;
}


NS_IMETHODIMP
nsDatabaseConnectionPgsql::GetStatusCode(PRInt32* aStatusCode)
{
  *aStatusCode = mStatusCode;
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::GetStatusMessage(char** aStatusMessage)
{
  *aStatusMessage = ToNewCString(mStatusMessage);
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::GetWindow(nsIDOMWindowInternal** aWindow)
{
  *aWindow = mWindow;
  NS_IF_ADDREF(*aWindow);

  return NS_OK;
}

NS_IMETHODIMP nsDatabaseConnectionPgsql::Init(nsIDOMWindowInternal* aWindow)
{
  mWindow = aWindow;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::Connect(const char* aHost, PRInt32 aPort, const char* aDatabase, const char* aUsername, const char* aPassword)
{
  if (mConnection) {
    mStatusCode = ERR_CONNECTED;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  char port[16];
  char options[]="";
  char tty[]="";

  PR_snprintf(port, 16, "%d", aPort);
  mConnection = PQsetdbLogin(aHost, port, options, tty, aDatabase, aUsername, aPassword);
  if (PQstatus(mConnection) == CONNECTION_BAD) {
    mConnection = nsnull;
    mStatusCode = ERR_CONNECT_FAILED;
    mStatusMessage.Assign(PQerrorMessage(mConnection));
  }
  else {
//    PQsetClientEncoding(mConnection, "UNICODE");
    static char query[] = "set client_encoding to 'UNICODE'";
    PGresult* result = PQexec(mConnection, query);
    PQclear(result);

    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
    mConnected = PR_TRUE;
  }

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::GetConnected(PRBool *aConnected)
{
  *aConnected = mConnected;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::Disconnect()
{
  if (! mConnection) {
    mStatusCode = ERR_NOT_CONNECTED;
    mStatusMessage.Assign("");
  }
  else {
    PQfinish(mConnection);

    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
    mConnected = PR_FALSE;
    mConnection = nsnull;
  }

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::GetRunning(PRBool *aRunning)
{
  *aRunning = mRunning;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::Exec(const PRUnichar* aQuery, nsIDatabaseRowList** _retval)
{
  *_retval = nsnull;

  if (! mWindow)
    return NS_ERROR_NOT_INITIALIZED;

  if (! mConnection) {
    mStatusCode = ERR_NOT_CONNECTED;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  if (mRunning) {
    mStatusCode = ERR_RUNNING;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  PR_Lock(mLock);

  mRunning = PR_TRUE;

  mCancelled = PR_FALSE;

  mWindow->SetCursor(NS_LITERAL_STRING("wait"));

  NotifyObservers(MSG_BEGIN_EXECUTION, PR_FALSE);

  PGresult* result;
  result = PQexec(mConnection, NS_ConvertUCS2toUTF8(aQuery).get());

  PRInt32 stat;
  stat = PQresultStatus(result);
  if ( (stat != PGRES_COMMAND_OK) && (stat != PGRES_TUPLES_OK) ) {
    mStatusCode = ERR_QUERY_FAILED;
    mStatusMessage.Assign(PQresultErrorMessage(result));
  }
  else {
    BuildRows(result, _retval);
    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
  }

  PQclear(result);

  NotifyObservers(MSG_END_EXECUTION, PR_FALSE);

  mWindow->SetCursor(NS_LITERAL_STRING("default"));

  mRunning = PR_FALSE;

  PR_Unlock(mLock);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::AsyncExec(const PRUnichar *aQuery, nsIDatabaseConnectionListener* aListener, nsISupports* aContext)
{
  if (! aListener)
    return NS_ERROR_NULL_POINTER;

  if (! mWindow)
    return NS_ERROR_NOT_INITIALIZED;

  if (! mConnection) {
    aListener->OnError(this, aContext, ERR_NOT_CONNECTED, "");
    return NS_OK;
  }

  if (mRunning) {
    aListener->OnError(this, aContext, ERR_RUNNING, "");
    return NS_OK;
  }

  mQuery.Assign(aQuery);

  mProxyObjectManager->GetProxyForObject(NS_UI_THREAD_EVENTQ,
                                         NS_GET_IID(nsIDatabaseConnectionListener),
                                         aListener,
                                         PROXY_SYNC | PROXY_ALWAYS,
                                         getter_AddRefs(mListener));

  mContext = aContext;

  PR_Lock(mLock);
  PR_NotifyCondVar(mCondVar);
  PR_Unlock(mLock);

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::GetCancelled(PRBool *aCancelled)
{
  *aCancelled = mCancelled;

  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::Stop()
{
  if (! mConnection) {
    mStatusCode = ERR_NOT_CONNECTED;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  if (! mRunning) {
    mStatusCode = ERR_NOT_RUNNING;
    mStatusMessage.Assign("");
    return NS_OK;
  }

  if (PQrequestCancel(mConnection)) {
    mStatusCode = ERR_OK;
    mStatusMessage.Assign("");
    mCancelled = PR_TRUE;
  }
  else {
    mStatusCode = ERR_STOP_FAILED;
    mStatusMessage.Assign("");
  }
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::AddObserver(nsIDatabaseConnectionObserver* aObserver)
{
  if (! mObservers) {
    nsresult rv;
    rv = NS_NewISupportsArray(getter_AddRefs(mObservers));
    if (NS_FAILED(rv))
      return rv;
  }
                
  mObservers->AppendElement(aObserver);
  
  return NS_OK;
}

NS_IMETHODIMP
nsDatabaseConnectionPgsql::RemoveObserver(nsIDatabaseConnectionObserver* aObserver)
{
  if (! mObservers)
    return NS_OK;
        
  mObservers->RemoveElement(aObserver);

  return NS_OK;
}


nsresult
nsDatabaseConnectionPgsql::BuildRows(PGresult* aResult, nsIDatabaseRowList** _retval)
{
#ifdef DEBUG
  PRIntervalTime startTime =  PR_IntervalNow();
#endif

  PRInt32 columnCount = PQnfields(aResult);

  nsCOMPtr<nsIDatabaseRowList> rowList;
  nsComponentManager::CreateInstance(kDatabaseRowListCID,
                                     nsnull,
                                     NS_GET_IID(nsIDatabaseRowList),
                                     getter_AddRefs(rowList));

  rowList->Init(columnCount);
  PRInt32 i;
  for (i = 0; i < columnCount; i++)
    rowList->SetColumnName(i, NS_ConvertUTF8toUCS2(PQfname(aResult, i)).get());
      
  for (i = 0; i < PQntuples(aResult); i++) {
    nsCOMPtr<nsIDatabaseRow> databaseRow;
    rowList->CreateRow(getter_AddRefs(databaseRow));
    PRInt32 j;
    for (j = 0; j < columnCount; j++)
      databaseRow->SetValue(j, NS_ConvertUTF8toUCS2(PQgetvalue(aResult, i, j)).get());
    rowList->AppendRow(databaseRow);
  }

#ifdef DEBUG
  PRIntervalTime endTime =  PR_IntervalNow();
  printf("Time spent with building rows: %d msec\n", PR_IntervalToMilliseconds(endTime - startTime));
#endif

  *_retval = rowList;
  NS_ADDREF(*_retval);
  
  return NS_OK;
}

nsresult
nsDatabaseConnectionPgsql::NotifyObservers(char aMsgType, PRBool aUseProxy)
{
  if (mObservers) {
     PRUint32 count;
     mObservers->Count(&count);
     for (PRUint32 i = 0; i < count; i++) {
       nsCOMPtr<nsIDatabaseConnectionObserver> observer;
       mObservers->QueryElementAt(i, NS_GET_IID(nsIDatabaseConnectionObserver), getter_AddRefs(observer));
       if (observer) {
         nsCOMPtr<nsIDatabaseConnectionObserver> realObserver;
         if (aUseProxy)
           mProxyObjectManager->GetProxyForObject(NS_UI_THREAD_EVENTQ,
                                                  NS_GET_IID(nsIDatabaseConnectionObserver),
                                                  observer,
                                                  PROXY_SYNC | PROXY_ALWAYS,
                                                  getter_AddRefs(realObserver));
         else
           realObserver = observer;
       
         if (aMsgType == MSG_BEGIN_EXECUTION)
           realObserver->OnBeginExecution();
         else if (aMsgType == MSG_END_EXECUTION)
           realObserver->OnEndExecution();
       }
     }
  }

  return NS_OK;
}
